#
# Copyright (C) 2019 Authlete, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
# either express or implied. See the License for the specific
# language governing permissions and limitations under the
# License.


import time
from authlete.django.handler.authorization_request_base_handler import AuthorizationRequestBaseHandler
from authlete.django.handler.claim_collector                    import ClaimCollector
from authlete.dto.authorization_action                          import AuthorizationAction
from authlete.dto.authorization_fail_reason                     import AuthorizationFailReason


class NoInteractionHandler(AuthorizationRequestBaseHandler):
    """Handler for the case where an authorization request should be processed without user interaction.

    A response from Authlete's /api/auth/authorization API contains an
    'action' response parameter. When the value of the response parameter
    is "NO_INTERACTION", the authorization request needs to be processed
    without user interaction. This class is a handler for the case.
    """


    def __init__(self, api, spi):
        """Constructor

        Args:
            api (authlete.api.AuthleteApi)
            spi (authlete.django.handler.spi.NoInteractionHandlerSpi)
        """

        super().__init__(api)
        self._spi = spi


    def handle(self, response):
        """Handle an authorization request without user interaction.

        This method calls Authlete's /api/auth/authorization/issue API or
        /api/auth/authorization/fail API.

        Args:
            response (authlete.dto.AuthorizationResponse)

        Returns:
            django.http.HttpResponse

        Raises:
            authlete.api.AuthleteApiException
        """

        spi = self._spi

        # If the value of the "action" parameter in the response from Authlete's
        # /api/auth/authorization API is not "NO_INTERACTION".
        if response.action != AuthorizationAction.NO_INTERACTION:
            # This handler does not handle other cases than NO_INTERACTION.
            return None

        # Check 1: User Authentication
        if self.__checkUserAuthentication() == False:
            # A user must have logged in.
            return self.authorizationFail(
                response.ticket, AuthorizationFailReason.NOT_LOGGED_IN)

        # Get the last time when the user was authenticated.
        authTime = spi.getUserAuthenticatedAt()

        # Check 2: Max Age
        if self.__checkMaxAge(response, authTime) == False:
            # The maximum authentication age has elapsed since the last time
            # when the user was authenticated.
            return self.authorizationFail(
                response.ticket, AuthorizationFailReason.EXCEEDS_MAX_AGE)

        # The subject (unique ID) of the current user.
        subject = spi.getUserSubject()

        # Check 3: Subject
        if self.__checkSubject(response, subject) == False:
            # The requested subject and that of the current user don't match.
            return self.authorizationFail(
                response.ticket, AuthorizationFailReason.DIFFERENT_SUBJECT)

        # Get the value of the "sub" claim. This is optional. When "sub" is None,
        # the value of "subject" will be used as the value of the "sub" claim.
        sub = spi.getSub()

        # Get the ACR that was satisfied when the current user was authenticated.
        acr = spi.getAcr()

        # Check 4: ACR
        if self.__checkAcr(response, acr) == False:
            # None of the requested ACRs is satisfied.
            return self.authorizationFail(
                response.ticket, AuthorizationFailReason.ACR_NOT_SATISFIED)

        # Collect claim values.
        claims = ClaimCollector(
            subject, response.claims, response.claimsLocales, spi).collect()

        # Properties to be associated with an access token and/or an authorization code.
        properties = spi.getProperties()

        # Scopes associated with an access token and/or an authorization code.
        # If the value returned from spi.getScopes() is not None, the scope set
        # replaces the scopes that were given by the original authorization request.
        scopes = spi.getScopes()

        # Issue tokens without user interaction.
        return self.authorizationIssue(
            response.ticket, subject, authTime, acr, claims, properties, scopes, sub)


    def __checkUserAuthentication(self):
        return self._spi.isUserAuthenticated()


    def __checkMaxAge(self, response, authTime):
        # Get the requested maximum authentication age.
        maxAge = response.maxAge

        # If no maximum authentication age is requested.
        if maxAge == 0:
            # No need to care about the maximum authentication age.
            return True

        # The time in seconds when the authentication expires.
        expiresAt = authTime + maxAge

        # If the authentication has not expired yet.
        if time.time() < expiresAt:
            # Not exceed.
            return True

        # Exceeded
        return False


    def __checkSubject(self, response, subject):
        # Get the requested subject.
        requestedSubject = response.subject

        # If no subject is requested.
        if requestedSubject is None:
            # Ne need to care about the subject.
            return True

        # If the requested subject matches that of the current user.
        if requestedSubject == subject:
            # The subjects match.
            return True

        # The subjects don't match.
        return False


    def __checkAcr(self, response, acr):
        # Get the list of requested ACRs.
        requestedAcrs = response.acrs

        # if no ACR is requested.
        if requestedAcrs is None or len(requestedAcrs) == 0:
            # No need to care about ACR.
            return True

        # For each requested ACR.
        for requestedAcr in requestedAcrs:
            if requestedAcr == acr:
                # OK. The ACR satisfied when the current user was authenticated
                # matches one of the requested ACRs.
                return True

        # If one of the requested ACRs must be satisfied.
        if response.acrEssential:
            # None of the requested ACRs is satisfied.
            return False

        # The ACR satisfied when the current user was authenticated does not
        # match any one of the requested ACRs, but the authorization request
        # from the client application did not request ACR as essential.
        # Therefore, it is not necessary to raise an error here.
        return True



#
# Copyright (C) 2019-2024 Authlete, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
# either express or implied. See the License for the specific
# language governing permissions and limitations under the
# License.


from json import dumps
from authlete.django.handler.base_request_handler import BaseRequestHandler
from authlete.django.web.response_utility         import ResponseUtility
from authlete.dto.token_fail_action               import TokenFailAction
from authlete.dto.token_fail_request              import TokenFailRequest
from authlete.dto.token_issue_action              import TokenIssueAction
from authlete.dto.token_issue_request             import TokenIssueRequest


class TokenRequestBaseHandler(BaseRequestHandler):
    """The base class for request handlers that are used in the implementation of a token endpoint.
    """


    def __init__(self, api):
        """Constructor

        Args:
            api (authlete.api.AuthleteApi)
        """

        super().__init__(api)


    def tokenIssue(self, ticket, subject, properties, headers=None):
        """Call /auth/token/issue API.

        Args:
            ticket (str)  : The ticket which has been issued previously from /auth/token API.
            subject (str) : The unique identifier of the resource owner.
            properties (list of authlete.dto.Property)
            headers (dict) : Additional HTTP headers.

        Returns:
            django.http.HttpResponse

        Raises:
            authlete.api.AuthleteApiException
        """

        # Call /api/auth/token/issue API.
        res = self.__callTokenIssue(ticket, subject, properties)

        # 'action' in the response denotes the next action which the
        # implementation of the token endpoint should take.
        action = res.action

        # The content of the response to the client application.
        content = res.responseContent

        if action == TokenIssueAction.INTERNAL_SERVER_ERROR:
            # 500 Internal Server Error
            return ResponseUtility.internalServerError(content, headers)
        elif action == TokenIssueAction.OK:
            # 200 OK
            return ResponseUtility.okJson(content, headers)
        else:
            # 500 Internal Server Error
            # The /api/auth/token/issue API returned an unknown action.
            return self.unknownAction('/api/auth/token/issue')


    def __callTokenIssue(self, ticket, subject, properties):
        # Prepare a request for /api/auth/token/issue API.
        req = TokenIssueRequest()
        req.ticket     = ticket
        req.subject    = subject
        req.properties = properties

        # Call /api/auth/token/issue API.
        return self.api.tokenIssue(req)


    def tokenFail(self, ticket, reason, headers=None):
        """Call /api/auth/authorization/fail API.

        Args:
            ticket (str) : The ticket which has been issued previously from /api/auth/token API.
            reason (authlete.dto.TokenFailReason) : The reason of the failure of the request.

        Returns:
            django.http.HttpResponse

        Raises:
            authlete.api.AuthleteApiException
        """

        # Call /api/auth/token/fail API.
        res = self.__callTokenFail(ticket, reason)

        # 'action' in the response denotes the next action which the
        # implementation of the token endpoint should take.
        action = res.action

        # The content of the response to the client application.
        content = res.responseContent

        if action == TokenFailAction.INTERNAL_SERVER_ERROR:
            # 500 Internal Server Error
            return ResponseUtility.internalServerError(content, headers)
        elif action == TokenFailAction.BAD_REQUEST:
            # 400 Bad Request
            return ResponseUtility.badRequest(content, headers)
        else:
            # 500 Internal Server Error
            # The /api/auth/token/fail API returned an unknown action.
            return self.unknownAction('/api/auth/token/fail')


    def __callTokenFail(self, ticket, reason):
        # Prepare a request for /api/auth/token/fail API.
        req = TokenFailRequest()
        req.ticket = ticket
        req.reason = reason

        # Call /api/auth/token/fail API.
        return self.api.tokenFail(req)



#
# Copyright (C) 2019 Authlete, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
# either express or implied. See the License for the specific
# language governing permissions and limitations under the
# License.


from authlete.django.handler.authorization_request_base_handler import AuthorizationRequestBaseHandler
from authlete.django.web.response_utility                       import ResponseUtility
from authlete.dto.authorization_action                          import AuthorizationAction


class AuthorizationRequestErrorHandler(AuthorizationRequestBaseHandler):
    """Handler for error cases of authorization reuqests.

    A response from Authlete's /api/auth/authorization API contains an "action"
    response parameter. When the value of the response parameter is neither
    "NO_INTERACTION" nor "INTERACTION", the authorization request should be
    handled as an error case. This class is a handler for such error cases.
    """


    def __init__(self):
        """Constructor"""
        super().__init__(None)


    def handle(self, response):
        """Handle an error case of an authorization request.

        This method returns None when response.action returns
        AuthorizationAction.INTERACTION or AuthorizationAction.NO_INTERACTION.
        In other cases, an instance of django.http.HttpResponse is returned.

        Args:
            response (authlete.dto.AuthorizationResponse)

        Returns:
            django.http.HttpResponse : An error response

        Raises:
            authlete.api.AuthleteApiException
        """

        # 'action' in the response denotes the next action which the
        # implementation of the authorization endpoint should take.
        action = response.action

        # The content of the response which should be returned to the
        # user agent. The format varies depending on the action.
        content = response.responseContent

        if action == AuthorizationAction.INTERNAL_SERVER_ERROR:
            # 500 Internal Server Error
            return ResponseUtility.internalServerError(content)
        elif action == AuthorizationAction.BAD_REQUEST:
            # 400 Bad Request
            return ResponseUtility.badRequest(content)
        elif action == AuthorizationAction.LOCATION:
            # 302 Found
            return ResponseUtility.location(content)
        elif action == AuthorizationAction.FORM:
            # 200 OK
            return ResponseUtility.okHtml(content)
        elif action == AuthorizationAction.INTERACTION:
            # This is not an error case. The implementation of the
            # authorization endpoint should show an authorization
            # page to the user.
            return None
        elif action == AuthorizationAction.NO_INTERACTION:
            # This is not an error case. The implementation of the
            # authorization endpoint should handle the authorization
            # request without user interaction.
            return None
        else:
            # 500 Internal Server Error
            # Authlete's /api/auth/authorization API returned an unknown action.
            return self.unknownAction('/api/auth/authorization')



#
# Copyright (C) 2019-2024 Authlete, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
# either express or implied. See the License for the specific
# language governing permissions and limitations under the
# License.


from abc import ABCMeta, abstractmethod


class TokenRequestHandlerSpi(metaclass=ABCMeta):
    """Service Provider Interface for TokenRequestHandler.

    An implementation of this interface needs to be given to the constructor of
    TokenRequestHandler.

    TokenRequestHandlerSpiAdapter is an empty implementation of this interface.
    """


    @abstractmethod
    def authenticateUser(self, username, password):
        """Authenticate a user.

        This method is called only when Resource Owner Password Credentials
        Grant (RFC 6749, 4.3) was used. Therefore, if you have no plan to
        support the flow, always return None. In most cases, you don't
        have to support the flow. RFC 6749 says "The authorization server
        should take special care when enabling this grant type and only
        allow it when other flows are not viable."

        Args:
            username (str) :  The value of the "username" request parameter of the token request.
            password (str) :  The value of the "password" request parameter of the token request.

        Returns:
            str : The subject (= unique identifier) of the authenticated user. None if not authenticated.
        """
        pass


    @abstractmethod
    def getProperties(self):
        """Get arbitrary key-value pairs to be associated with an access token.

        Properties returned from this method will appear as top-level entries
        (unless they are marked as hidden) in a JSON response from the
        authorization server as shown in RFC 6749, 5.1.

        Returns:
            list : list of authlete.dto.Property.
        """
        pass


    @abstractmethod
    def tokenExchange(self, tokenResponse):
        """Handle a token exchange request.

        This method is called when the grant type of the token request is
        'urn:ietf:params:oauth:grant-type:token-exchange'. The grant type is
        defined in 'RFC 8693: OAuth 2.0 Token Exchange'.

        RFC 8693 is very flexible. In other words, the specification does not
        define details that are necessary for secure token exchange. Therefore,
        implementations have to complement the specifications with their own
        rules.

        The argument passed to this method is an instance of
        'authlete.dto.TokenResponse' that represents a response from Authlete's
        /auth/token API. The instance contains information about the token
        exchange request such as the value of the 'subject_token' request
        parameter. Implementations of this 'tokenExchange' method are supposed
        to (1) validate the information based on their own rules, (2) generate
        a token (e.g. an access token) using the information, and (3) prepare
        a token response in the JSON format that conforms to Section 2.2 of
        RFC 8693.

        Authlete's /auth/token API performs validation of token exchange
        requests to some extent. Therefore, authorization server implementations
        don't have to repeat the same validation steps. See the JavaDoc of the
        'TokenResponse' class for details about the validation steps.
        ( https://authlete.github.io/authlete-java-common/ )

        This method should return a django.http.HttpResponse instance that
        conforms to RFC 8693. If this method returns 'None', 'TokenRequestHandler'
        will generate an HTTP response with 400 Bad Request and
        {"error":"unsupported_grant_type"}.

        NOTE: Token Exchange is supported by Authlete 2.3 and newer versions.
        If the Authlete server of your system is older than version 2.3, the
        grant type 'urn:ietf:params:oauth:grant-type:token-exchange' is not
        supported and so this method is never called.

        Args:
            tokenResponse (authlete.dto.TokenResponse)

        Returns:
            django.http.HttpResponse : A response from the token endpoint.
        """
        pass


    @abstractmethod
    def jwtBearer(self, tokenResponse):
        """Handle a token request that uses the grant type
        'urn:ietf:params:oauth:grant-type:jwt-bearer'.

        This method is called when the grant type of the token request is
        'urn:ietf:params:oauth:grant-type:jwt-bearer'. The grant type is defined
        in 'RFC 7523: JSON Web Token (JWT) Profile for OAuth 2.0 Client
        Authentication and Authorization Grants'.

        The grant type utilizes a JWT as an authorization grant, but the
        specification does not define details about how the JWT is generated by
        whom. As a result, it is not defined in the specification how to obtain
        the key whereby to verify the signature of the JWT. Therefore, each
        deployment has to define their own rules which are necessary to determine
        the key for signature verification.

        The argument passed to this method is an instance of
        'authlete.dto.TokenResponse' that represents a response from Authlete's
        /auth/token API. The instance contains information about the token request
        such as the value of the 'assertion' request parameter. Implementations
        of this method are supposed to (1) validate the authorization grant
        (= the JWT specified by the 'assertion' request parameter), (2) generate
        an access token, and (3) prepare a token response in the JSON format that
        conforms to RFC 6749.

        Authlete's /auth/token API performs validation of token requests to some
        extent. Therefore, authorization server implementations don't have to
        repeat the same validation steps. Basically, what implementations have to
        do is to verify the signature of the JWT. See the JavaDoc of the
        'TokenResponse' class for details about the validation steps.
        ( https://authlete.github.io/authlete-java-common/ )

        This method should return a django.http.HttpResponse instance that
        conforms to RFC 6749. If this method returns 'None', 'TokenRequestHandler'
        will generate an HTTP response with 400 Bad Request and
        {"error":"unsupported_grant_type"}.

        NOTE: JWT Authorization Grant is supported by Authlete 2.3 and newer
        versions. If the Authlete server of your system is older than version 2.3,
        the grant type 'urn:ietf:params:oauth:grant-type:jwt-bearer' is not
        supported and so this method is never called.

        Args:
            tokenResponse (authlete.dto.TokenResponse)

        Returns:
            django.http.HttpResponse : A response from the token endpoint.
        """
        pass



#
# Copyright (C) 2019 Authlete, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
# either express or implied. See the License for the specific
# language governing permissions and limitations under the
# License.


import base64
import re


class BasicCredentials(object):
    BASIC_PATTERN = re.compile('^Basic *(?P<parameter>[^ ]+) *$', re.I)


    def __init__(self, userId, password):
        self._userId             = userId
        self._password           = password
        self._formattedParameter = self.__class__.__formatParameter(userId, password)
        self._formatted          = 'Basic {}'.format(self._formattedParameter)


    @property
    def userId(self):
        return self._userId


    @property
    def password(self):
        return self._password


    @property
    def formatted(self):
        return self._formatted


    @property
    def formattedParameter(self):
        return self._formattedParameter


    @classmethod
    def __formatParameter(cls, userId, password):
        if userId is None:
            userId = ''

        if password is None:
            password = ''

        plain = '{}:{}'.format(userId, password)

        return cls.__b64encode(plain)


    @classmethod
    def __bytes_io(cls, input, func):
        return func(input.encode('utf-8')).decode('utf-8')


    @classmethod
    def __b64encode(cls, input):
        return cls.__bytes_io(input, base64.b64encode)


    @classmethod
    def __b64decode(cls, input):
        return cls.__bytes_io(input, base64.b64decode)


    @classmethod
    def parse(cls, input):
        """Create a BasicCredentials instance from the given string.

        The format of the input string should be "Basic {base64-encoded-string}".
        If the given string is None or it does not match the pattern, a
        BasicCredentials instance whose userId and password are both None is
        returned.

        Args:
            input (str): The value of the Authorization header.

        Returns:
            authlete.django.web.BasicCredentials
        """

        if input is None:
            # userId = None, password = None
            return BasicCredentials(None, None)

        # Expecting the input matches "Basic {base64string}"
        mo = cls.BASIC_PATTERN.match(input)

        # If the input string does not match the pattern.
        if mo is None:
            # userId = None, password = None
            return BasicCredentials(None, None)

        # BASE64-encoded "userId:password"
        base64string = mo.group('parameter')

        # Build a BasicCredentials instance from the BASE64 string.
        return cls.__buildFromParameter(base64string)


    @classmethod
    def __buildFromParameter(cls, base64string):
        if base64string is None or len(base64string) == 0:
            # userId = None, password = None
            return BasicCredentials(None, None)

        # Decode the BASE64 string
        plain = cls.__b64decode(base64string)

        # Split "userId:password" into "userId" and "password"
        elements = plain.split(':', 2)
        count    = len(elements)

        userId   = None
        password = None

        # User ID
        if 1 <= count:
            userId = elements[0]

        # Password
        if 2 <= count:
            password = elements[1]

        return BasicCredentials(userId, password)



#
# Copyright (C) 2019 Authlete, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
# either express or implied. See the License for the specific
# language governing permissions and limitations under the
# License.


import unittest
from authlete.django.web.basic_credentials import BasicCredentials


class TestBasicCredentials(unittest.TestCase):
    def setUp(self):
        self.userId       = "user"
        self.password     = "password"
        self.base64       = "dXNlcjpwYXNzd29yZA=="
        self.basic_base64 = "Basic {}".format(self.base64)


    def test_001(self):
        actual = BasicCredentials(self.userId, self.password).formatted
        self.assertEqual(actual, self.basic_base64)


    def test_002(self):
        actual = BasicCredentials(self.userId, self.password).formattedParameter
        self.assertEqual(actual, self.base64)


    def test_003(self):
        credentials = BasicCredentials.parse(self.basic_base64)
        actual      = credentials.formatted

        self.assertEqual(actual, self.basic_base64)


    def test_004(self):
        header      = "Dummy {}".format(self.base64)
        credentials = BasicCredentials.parse(header)

        self.assertIsNone(credentials.userId)
        self.assertIsNone(credentials.password)



#
# Copyright (C) 2019 Authlete, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
# either express or implied. See the License for the specific
# language governing permissions and limitations under the
# License.


class ClaimCollector(object):
    def __init__(self, subject, claimNames, claimLocales, claimProvider):
        """Constructor

        Args:
            subject (str)       : The subject (= unique identifier) of the user.
            claimNames (list)   : list of str. Claim names.
            claimLocales (list) : list of str. Claim locales.
            claimProvider (authlete.django.handler.spi.ClaimProviderSpi)
        """

        self._subject       = subject
        self._claimNames    = claimNames
        self._claimLocales  = self.__normalizeClaimLocales(claimLocales)
        self._claimProvider = claimProvider


    def __normalizeClaimLocales(self, claimLocales):
        if claimLocales is None or len(claimLocales) == 0:
            return None

        # From "5.2. Claims Languages and Scripts" in OpenID Connect Core 1.0
        #
        #   However, since BCP47 language tag values are case insensitive,
        #   implementations SHOULD interpret the language tag values supplied
        #   in a case insensitive manner.
        #

        localeList = list()
        localeSet  = set()
        for claimLocale in claimLocales:
            if claimLocale is None or len(claimLocale) == 0:
                continue

            locale = claimLocale.lower()
            if locale in localeSet:
                continue

            localeSet.add(locale)
            localeList.append(locale)

        if len(localeList) == 0:
            return None

        return localeList


    def collect(self):
        claimNames = self._claimNames

        if claimNames is None or len(claimNames) == 0:
            return None

        # Pairs of claim name and its value.
        collectedClaims = {}

        # For each required claim.
        for claimName in claimNames:
            if claimName is None:
                continue

            # Split the claim name into the name part and the language tag part.
            elements = claimName.split('#', 2)
            name = elements[0]
            tag  = None if len(elements) != 2 else elements[1]

            # If the name part is empty.
            if name is None or len(name) == 0:
                continue

            # Get the value of the claim.
            value = self.__getClaimValue(name, tag)

            # If the value of the claim was not obtained.
            if value is None:
                continue

            # Just for an edge case where claimName ends with '#'. e.g. 'family_name#'
            if tag is None or len(tag) == 0:
                claimName = name

            # Add the pair of the claim name and its value.
            collectedClaims[claimName] = value

        # If no claim value has been obtained.
        if len(collectedClaims) == 0:
            return None

        return collectedClaims


    def __getClaimValue(self, name, tag):
        provider = self._claimProvider
        subject  = self._subject

        # If a language tag is explicitly appended.
        if tag is not None and len(tag) != 0:
            # Get the claim value of the claim with the specific language tag.
            return provider.getUserClaimValue(subject, name, tag)

        # If claim locales are not specified by the 'claims_locales' parameter.
        if self._claimLocales is None:
            # Get the claim value of the claim without any language tag.
            return provider.getUserClaimValue(subject, name, None)

        # For each claim locale. They are ordered by preference.
        for locale in self._claimLocales:
            # Try to get the claim value with the claim locale.
            value = provider.getUserClaimValue(subject, name, locale)

            # If the claim value was obtained.
            if value is not None:
                return value

        # The last resort. Try to get the claim value without any language tag.
        return provider.getUserClaimValue(subject, name, None)



#
# Copyright (C) 2019 Authlete, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
# either express or implied. See the License for the specific
# language governing permissions and limitations under the
# License.


from abc import abstractmethod
from .user_claim_provider_spi import UserClaimProviderSpi


class AuthorizationRequestHandlerSpi(UserClaimProviderSpi):
    """The base interface for Service Provider Interfaces for authorization request handlers.

    Common methods for NoInteractionHandlerSpi and AuthorizationRequestDecisionHandlerSpi.
    """


    @abstractmethod
    def getUserAuthenticatedAt(self):
        """Get the time when the user was authenticated.

        Returns:
            int :
                The time when the current user was authenticated. The number of
                seconds since the Unix epoch. 0 means that the time is unknown.
        """
        pass


    @abstractmethod
    def getUserSubject(self):
        """Get the subject (= unique identifier) of the user.

        It must consist of only ASCII letters and its length must not exceed 100.

        Returns:
            str : The subject of the user.
        """
        pass


    @abstractmethod
    def getSub(self):
        """Get the value of the "sub" claim that will be embedded in an ID token.

        If this method returns None, the value returned from getUserSubject() will
        be used.

        The main purpose of this method is to hide the actual value of the subject
        from client applications.

        Returns:
            str : The value of the "sub" claim.
        """
        pass


    @abstractmethod
    def getAcr(self):
        """Get the authentication context class referece (ACR) that was satisfied when the user was authenticated.

        The value returned from this method has an important meaning only when the
        "acr" claim is requested as an essential claim. See OIDC Core, 5.1.1.1.
        for details.

        Returns:
            str : The ACR that was satisfied when the user was authenticated.
        """
        pass


    @abstractmethod
    def getProperties(self):
        """Get arbitrary key-value pairs to be associated with an access token and/or an authorization code.

        Properties returned from this method will appear as top-level entries
        (unless they are marked as hidden) in a JSON response from the
        authorization server as shown in RFC 6749, 5.1.

        Returns:
            list : list of authlete.dto.Property.
        """
        pass


    @abstractmethod
    def getScopes(self):
        """Get scopes to be associated with an access token and/or an authorization code.

        If None is returned, the scopes specified in the original authorization
        request from the client application are used. In other cases, the
        specified scopes by this method will replace the original scopes.

        Even scopes that are not included in the original authorization request
        can be specified. However, as an exception, the "openid" scope is ignored
        on Authlete server side if it is not included in the original request.
        It is because the existence of the "openid" scope considerably changes
        the validation steps and because adding "openid" triggers generation of
        an ID token (although the client application has not requested it) and
        the behavior is a major violation against the specification.

        If you add "offline_access" scope although it is not included in the
        original request, keep in mind that the specification requires explicit
        consent from the user forthe scope (OIDC Core, 11). When "offline_access"
        is included in the original authorization request, the current
        implementation of Authlete's /api/auth/authorization API checks whether
        the authorization request has come along with the "prompt" request
        parameter and its value includes "consent". However, note that the
        implementation of Authlete's /api/auth/authorization/issue API does not
        perform the same validation even if the "offline_access" scope is newly
        added via this method.

        Returns:
            list : list of str. Scope names.
        """
        pass



#
# Copyright (C) 2019-2024 Authlete, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
# either express or implied. See the License for the specific
# language governing permissions and limitations under the
# License.


from django.http import HttpResponse


class ResponseUtility(object):
    @classmethod
    def okJson(cls, content, headers=None):
        # 200 OK, application/json;charset=UTF-8
        return cls.__json(200, content, headers)


    @classmethod
    def okJavaScript(cls, content, headers=None):
        # 200 OK, application/javascript;charset=UTF-8
        return cls.__javascript(200, content, headers)


    @classmethod
    def okJwt(cls, content, headers=None):
        # 200 OK, application/jwt
        return cls.__jwt(200, content, headers)


    @classmethod
    def okHtml(cls, content, headers=None):
        # 200 OK, text/html;charset=UTF-8
        return cls.__html(200, content, headers)


    @classmethod
    def entityStatement(cls, content, headers=None):
        # 200 OK, application/entity-statement+jwt
        return cls.__entityStatement(200, content, headers)


    @classmethod
    def created(cls, content, headers=None):
        # 201 Created, application/json;charset=UTF-8
        return cls.__json(201, content, headers)


    @classmethod
    def acceptedJson(cls, content, headers=None):
        # 202 Accepted, application/json;charset=UTF-8
        return cls.__json(202, content, headers)


    @classmethod
    def acceptedJwt(cls, content, headers=None):
        # 202 Accepted, application/jwt
        return cls.__jwt(202, content, headers)


    @classmethod
    def noContent(cls, headers=None):
        # 204 No Content
        return cls.__common(HttpResponse(status=204), headers)


    @classmethod
    def location(cls, location, headers=None):
        # 302 Found with a Location header.
        response = cls.__common(HttpResponse(status=302), headers)
        response['Location'] = location

        return response


    @classmethod
    def badRequest(cls, content, headers=None):
        # 400 Bad Request, application/json;charset=UTF-8
        return cls.__json(400, content, headers)


    @classmethod
    def unauthorized(cls, challenge, content=None, headers=None):
        # 401 Unauthorized with a WWW-Authenticate header
        return cls.wwwAuthenticate(401, challenge, content, headers)


    @classmethod
    def forbidden(cls, content, headers=None):
        # 403 Forbidden, application/json;charset=UTF-8
        return cls.__json(403, content, headers)


    @classmethod
    def notFound(cls, content, headers=None):
        # 404 Not Found, application/json;charset=UTF-8
        return cls.__json(404, content, headers)


    @classmethod
    def tooLarge(cls, content, headers=None):
        # 413 Too Large, application/json;charset=UTF-8
        return cls.__json(413, content, headers)


    @classmethod
    def internalServerError(cls, content, headers=None):
        # 500 Internal Server Error, application/json;charset=UTF-8
        return cls.__json(500, content, headers)


    @classmethod
    def wwwAuthenticate(cls, status, challenge, content=None, headers=None):
        if content is None:
            response = cls.__common(HttpResponse(status=status), headers)
        else:
            response = cls.__json(status, content, headers)

        response['WWW-Authenticate'] = challenge

        return response


    @classmethod
    def __response(cls, status, content, headers, content_type, charset='UTF-8'):
        response = HttpResponse(
            status=status, content=content,
            content_type=content_type, charset=charset)

        return cls.__common(response, headers)


    @classmethod
    def __common(cls, response, headers):
        response['Cache-Control'] = 'no-store'
        response['Pragma']        = 'no-cache'

        if headers is not None:
            for name, value in headers.items():
                response[ name ] = value

        return response


    @classmethod
    def __json(cls, status, content, headers):
        return cls.__response(status, content, headers, 'application/json')


    @classmethod
    def __javascript(cls, status, content, headers):
        return cls.__response(status, content, headers, 'application/javascript')


    @classmethod
    def __jwt(cls, status, content, headers):
        return cls.__response(status, content, headers, 'application/jwt', None)


    @classmethod
    def __html(cls, status, content, headers):
        return cls.__response(status, content, headers, 'text/html')


    @classmethod
    def __entityStatement(cls, status, content, headers):
        return cls.__response(status, content, headers, 'application/entity-statement+jwt')



#
# Copyright (C) 2019 Authlete, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
# either express or implied. See the License for the specific
# language governing permissions and limitations under the
# License.


from abc import abstractmethod
from .user_claim_provider_spi import UserClaimProviderSpi


class UserInfoRequestHandlerSpi(UserClaimProviderSpi):
    """Service Provider Interface for UserInfoRequestHandler.

    An implementation of this interface needs to be given to the constructor of
    UserInfoRequestHandler.

    UserInfoRequestHandlerSpiAdapter is an empty implementation of this interface.
    """


    @abstractmethod
    def getSub(self):
        """Get the value of the "sub" claim that will be embedded in the response from the userinfo endpoint.

        If this method returns None, the subject associated with the access
        token (which was presented by the client application at the userinfo
        endpoint) will be used as the value of the "sub" claim.

        Returns:
            str : The value of the "sub" claim.
        """
        pass



#
# Copyright (C) 2024 Authlete, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
# either express or implied. See the License for the specific
# language governing permissions and limitations under the
# License.


from authlete.django.handler.base_request_handler import BaseRequestHandler
from authlete.django.web.response_utility         import ResponseUtility
from authlete.dto.federation_registration_action  import FederationRegistrationAction


class FederationRegistrationRequestHandler(BaseRequestHandler):
    """Handler for requests to a federation registration endpoint.
    """


    def __init__(self, api):
        """Constructor

        Args:
            api (authlete.api.AuthleteApi)
        """

        super().__init__(api)


    def handle(self, request):
        """Handle a request to a federation registration endpoint.

        This method calls Authlete's /federation/registration API.

        Args:
            request (authlete.dto.FederationRegistrationRequest)

        Returns:
            django.http.HttpResponse

        Raises:
            authlete.api.AuthleteApiException
        """

        # Call Authlete's /federation/registration API.
        res = self.api.federationRegistration(request)

        # 'action' in the response denotes the next action which
        # the implementation of the endpoint should take.
        action = res.action

        # The content of the response.
        content = res.responseContent

        if action == FederationRegistrationAction.OK:
            # 200 OK; application/entity-statement+jwt
            return ResponseUtility.entityStatement(content)
        elif action == FederationRegistrationAction.BAD_REQUEST:
            # 400 Bad Request
            return ResponseUtility.badRequest(content)
        elif action == FederationRegistrationAction.NOT_FOUND:
            # 404 Not Found
            return ResponseUtility.notFound(content)
        elif action == FederationRegistrationAction.INTERNAL_SERVER_ERROR:
            # 500 Internal Server Error
            return ResponseUtility.internalServerError(content)
        else:
            # 500 Internal Server Error
            # /federation/registration API returns an unknown action.
            return self.unknownAction('/federation/registration')



#
# Copyright (C) 2019-2024 Authlete, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
# either express or implied. See the License for the specific
# language governing permissions and limitations under the
# License.


from authlete.django.handler.base_request_handler import BaseRequestHandler
from authlete.django.web.response_utility         import ResponseUtility
from authlete.dto.service_configuration_request   import ServiceConfigurationRequest


class ConfigurationRequestHandler(BaseRequestHandler):
    """Handler for requests to a configuration endpoint.

    An OpenID Provider that supports "OpenID Connect Discovery 1.0" provides an
    endpoint that returns its configuration information in JSON format. Details
    about the format are described in "3. OpenId Provider Metadata" of
    "OpenID Connect Discovery 1.0".

    Note that the URI of an OpenID Provider configuration endpoint is defined
    in "4.1. OpenID Provider Configuration Request". In short, the URI must be
    "{Issue-Identifier}/.well-known/openid-configuration".

    "{Issuer-Identifier}" is a URL that identifies an OpenID Provider. For
    example, "https://example.com". For details about Issuer Identifier, see
    the description about the "issuer" metadata defined in "3. OpenID Provider
    Metadata" (OpenID Connecto Discovery 1.0) and the "iss" claim in
    "2. ID Token" (OpenID Connect Core 1.0).
    """


    def __init__(self, api):
        """Constructor

        Args:
            api (authlete.api.AuthleteApi)
        """

        super().__init__(api)


    def handle(self, request, pretty=True):
        """Handle a request to a configuration endpoint.

        This method calls Authlete's /service/configuration API.

        Args:
            request (django.http.HttpRequest)
            pretty (bool)

        Returns:
            django.http.HttpResponse

        Raises:
            authlete.api.AuthleteApiException
        """

        # Call Authlete's /service/configuration API. The API returns
        # JSON that complies with OpenID Connect Discovery 1.0.
        req = ServiceConfigurationRequest()
        req.pretty = pretty

        jsn = self.api.getServiceConfiguration(req)

        # 200 OK, application/json;charset=UTF-8
        return ResponseUtility.okJson(jsn)



#
# Copyright (C) 2019 Authlete, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
# either express or implied. See the License for the specific
# language governing permissions and limitations under the
# License.


from authlete.django.handler.base_request_handler import BaseRequestHandler
from authlete.django.web.request_utility          import RequestUtility
from authlete.django.web.response_utility         import ResponseUtility
from authlete.dto.standard_introspection_action   import StandardIntrospectionAction
from authlete.dto.standard_introspection_request  import StandardIntrospectionRequest


class IntrospectionRequestHandler(BaseRequestHandler):
    """Handler for requests to an introspection endpoint (RFC 7662).
    """


    def __init__(self, api):
        """Constructor

        Args:
            api (authlete.api.AuthleteApi)
        """

        super().__init__(api)


    def handle(self, request):
        """Handle an introspection request.

        This method calls Authlete's /api/auth/introspection/starndard API.

        Args:
            request (django.http.HttpRequest)

        Returns:
            django.http.HttpResponse

        Raises:
            authlete.api.AuthleteApiException
        """

        # Request parameters in the request body.
        params = RequestUtility.extractRequestBody(request)

        # Call Authlete's /api/auth/introspection/standard API.
        res = self.__callStandardIntrospectionApi(params)

        # 'action' in the response denotes the next action which the
        # implementation of the introspection endpoint should take.
        action = res.action

        # The content of the response to the client application.
        content = res.responseContent

        if action == StandardIntrospectionAction.INTERNAL_SERVER_ERROR:
            # 500 Internal Server Error
            return ResponseUtility.internalServerError(content)
        elif action == StandardIntrospectionAction.BAD_REQUEST:
            # 400 Bad Request
            return ResponseUtility.badRequest(content)
        elif action == StandardIntrospectionAction.OK:
            # 200 OK
            return ResponseUtility.okJson(content)
        else:
            # 500 Internal Server Error
            # /api/auth/introspection/standard API returns an unknown action.
            return self.unknownAction('/api/auth/introspection/standard')


    def __callStandardIntrospectionApi(self, parameters):
        if parameters is None:
            # Authlete returns different error coes for None and an empty
            # string. None is regarded as a caller's error. An empty string
            # is regarded as a client application's error.
            parameters = ''

        # Create a request for /api/auth/introspection/standard API.
        req = StandardIntrospectionRequest()
        req.parameters = parameters

        # Call /api/auth/introspection/standard API.
        return self.api.standardIntrospection(req)



#
# Copyright (C) 2019 Authlete, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
# either express or implied. See the License for the specific
# language governing permissions and limitations under the
# License.


from json import dumps
from authlete.django.handler.base_request_handler import BaseRequestHandler
from authlete.django.web.response_utility         import ResponseUtility
from authlete.dto.authorization_fail_action       import AuthorizationFailAction
from authlete.dto.authorization_fail_request      import AuthorizationFailRequest
from authlete.dto.authorization_issue_action      import AuthorizationIssueAction
from authlete.dto.authorization_issue_request     import AuthorizationIssueRequest


class AuthorizationRequestBaseHandler(BaseRequestHandler):
    """The base class for request handlers that are used in the implementation of an authorization endpoint.
    """


    def __init__(self, api):
        """Constructor

        Args:
            api (authlete.api.AuthleteApi)
        """

        super().__init__(api)


    def authorizationIssue(self, ticket, subject, authTime, acr, claims, properties, scopes, sub):
        """Call /api/auth/authorization/issue API.

        Args:
            ticket (str)      : The ticket which has been issued previously from /api/auth/authorization API.
            subject (str)     : The unique identifier of the user who has granted permissions to the client.
            authTime (int)    : The time at which the user was authenticated. Seconds since the Unix epoch.
            acr (str)         : The Authentication Context Class Reference performed for user authentication.
            claims (dict)     : Claims about the user.
            properties (list) : list of authlete.dto.Property. Arbitrary properties to be associated with tokens.
            scopes (list)     : list of str. Scopes to be associated with tokens.
            sub (str)         : The value of the "sub" claim in an ID token.

        Returns:
            django.http.HttpResponse

        Raises:
            authlete.api.AuthleteApiException
        """

        # Call /api/auth/authorization API.
        res = self.__callAuthorizationIssue(
            ticket, subject, authTime, acr, claims, properties, scopes, sub)

        # 'action' in the response denotes the next action which the
        # implementation of the authorization endpoint should take.
        action = res.action

        # The content of the response to the user agent. The format
        # of the content varies depending on the action.
        content = res.responseContent

        if action == AuthorizationIssueAction.INTERNAL_SERVER_ERROR:
            # 500 Internal Server Error
            return ResponseUtility.internalServerError(content)
        elif action == AuthorizationIssueAction.BAD_REQUEST:
            # 400 Bad Request
            return ResponseUtility.badRequest(content)
        elif action == AuthorizationIssueAction.LOCATION:
            # 302 Found
            return ResponseUtility.location(content)
        elif action == AuthorizationIssueAction.FORM:
            # 200 OK
            return ResponseUtility.okHtml(content)
        else:
            # 500 Internal Server Error
            # The /api/auth/authorization/issue API returned an unknown action.
            return self.unknownAction('/api/auth/authorization/issue')


    def __callAuthorizationIssue(self, ticket, subject, authTime, acr, claims, properties, scopes, sub):
        # Prepare a request for /api/auth/authorization/issue API.
        req = AuthorizationIssueRequest()
        req.ticket     = ticket
        req.subject    = subject
        req.authTime   = authTime
        req.acr        = acr
        req.properties = properties
        req.scopes     = scopes
        req.sub        = sub

        if claims is not None:
            req.claims = dumps(claims)

        # Call /api/auth/authorization/issue API.
        return self.api.authorizationIssue(req)


    def authorizationFail(self, ticket, reason):
        """Call /api/auth/authorization/fail API.

        Args:
            ticket (str) : The ticket which has been issued previously from /api/auth/authorization API.
            reason (authlete.dto.AuthorizationFailReason) : The reason of the failure of the request.

        Returns:
            django.http.HttpResponse

        Raises:
            authlete.api.AuthleteApiException
        """

        # Call /api/auth/authorization/fail API.
        res = self.__callAuthorizationFail(ticket, reason)

        # 'action' in the response denotes the next action which the
        # implementation of the authorization endpoint should take.
        action = res.action

        # The content of the response to the user agent. The format
        # of the content varies depending on the action.
        content = res.responseContent

        if action == AuthorizationFailAction.INTERNAL_SERVER_ERROR:
            # 500 Internal Server Error
            return ResponseUtility.internalServerError(content)
        elif action == AuthorizationFailAction.BAD_REQUEST:
            # 400 Bad Request
            return ResponseUtility.badRequest(content)
        elif action == AuthorizationFailAction.LOCATION:
            # 302 Found
            return ResponseUtility.location(content)
        elif action == AuthorizationFailAction.FORM:
            # 200 OK
            return ResponseUtility.okHtml(content)
        else:
            # 500 Internal Server Error
            # The /api/auth/authorization/fail API returned an unknown action.
            return self.unknownAction('/api/auth/authorization/fail')


    def __callAuthorizationFail(self, ticket, reason):
        # Prepare a request for /api/auth/authorization/fail API.
        req = AuthorizationFailRequest()
        req.ticket = ticket
        req.reason = reason

        # Call /api/auth/authorization/fail API.
        return self.api.authorizationFail(req)



#
# Copyright (C) 2019 Authlete, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
# either express or implied. See the License for the specific
# language governing permissions and limitations under the
# License.


from .authorization_request_handler_spi_adapter import AuthorizationRequestHandlerSpiAdapter
from .no_interaction_handler_spi                import NoInteractionHandlerSpi


class NoInteractionHandlerSpiAdapter(
    AuthorizationRequestHandlerSpiAdapter, NoInteractionHandlerSpi):
    """An empty implementation of NoInteractionHandlerSpi.
    """


    def isUserAuthenticated(self):
        return False



#
# Copyright (C) 2019 Authlete, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
# either express or implied. See the License for the specific
# language governing permissions and limitations under the
# License.


from abc import abstractmethod
from .authorization_request_handler_spi import AuthorizationRequestHandlerSpi


class NoInteractionHandlerSpi(AuthorizationRequestHandlerSpi):
    """Service Provider Interface for NoInteractionHandler.

    An implementation of this interface needs to be given to the constructor of
    NoInteractionHandler.
    """


    @abstractmethod
    def isUserAuthenticated(self):
        """Check whether the user has already logged in or not.

        Returns:
            bool : True if the user has already logged in.
        """
        pass



#
# Copyright (C) 2019 Authlete, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
# either express or implied. See the License for the specific
# language governing permissions and limitations under the
# License.


from authlete.django.handler.authorization_request_base_handler import AuthorizationRequestBaseHandler
from authlete.django.handler.claim_collector                    import ClaimCollector
from authlete.dto.authorization_fail_reason                     import AuthorizationFailReason


class AuthorizationRequestDecisionHandler(AuthorizationRequestBaseHandler):
    """Handler for the user's decision on the authorization request.

    An authorization endpoint returns an authorization page (HTML) to the user,
    and the user will either authorize or deny the authorization request. The
    handle() method handles the decision and calls Authlete's
    /api/auth/authorization/issue API or /api/auth/authorization/fail API.
    """


    def __init__(self, api, spi):
        """Constructor

        Args:
            api (authlete.api.AuthleteApi)
            spi (authlete.django.handler.spi.AuthorizationRequestDecisionHandlerSpi)
        """

        super().__init__(api)
        self._spi = spi


    def handle(self, ticket, claimNames, claimLocales):
        """Handle the user's decision on the authorization request.

        Args:
            ticket (str)        : The ticket which has been issued previously by /api/auth/authorization API.
            claimNames (list)   : list of str. The value of 'claims' in the response from /api/auth/authorization API.
            claimLocales (list) : list of str. The value of 'claimsLocales' in the response from /api/auth/authorization API.

        Returns:
            django.http.HttpResponse

        Raises:
            authlete.api.AuthleteApiException
        """

        spi = self._spi

        # If the user did not grant authorization to the client application.
        if spi.isClientAuthorized() == False:
            # The user denied the authorization request.
            return self.authorizationFail(ticket, AuthorizationFailReason.DENIED)

        # The subject (= unique idnetifier) of the user.
        subject = spi.getUserSubject()

        # If the subject of the user is not available.
        if subject is None:
            # The user is not authenticated.
            return self.authorizationFail(ticket, AuthorizationFailReason.NOT_AUTHENTICATED)

        # Get the value of the "sub" claim. This is optional. When "sub" is None,
        # the value of "subject" will be used as the value of the "sub" claim.
        sub = spi.getSub()

        # The time when the user was authenticated.
        authTime = spi.getUserAuthenticatedAt()

        # The ACR (Authentication Context Class Reference) of the user authentication.
        acr = spi.getAcr()

        # Collect claim values.
        claims = ClaimCollector(subject, claimNames, claimLocales, spi).collect()

        # Properties to be associated with an access token and/or an authorization code.
        properties = spi.getProperties()

        # Scopes associated with an access token and/or an authorization code.
        # If the value returned from spi.getScopes() is not None, the scope set
        # replaces the scopes that were given by the original authorization request.
        scopes = spi.getScopes()

        # Issue required tokens by calling /api/auth/authorization/issue API.
        return self.authorizationIssue(
            ticket, subject, authTime, acr, claims, properties, scopes, sub)



#
# Copyright (C) 2019-2024 Authlete, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
# either express or implied. See the License for the specific
# language governing permissions and limitations under the
# License.


from json import dumps
from authlete.django.handler.base_request_handler import BaseRequestHandler
from authlete.django.handler.claim_collector      import ClaimCollector
from authlete.django.web.request_utility          import RequestUtility
from authlete.django.web.response_utility         import ResponseUtility
from authlete.dto.userinfo_action                 import UserInfoAction
from authlete.dto.userinfo_issue_action           import UserInfoIssueAction
from authlete.dto.userinfo_issue_request          import UserInfoIssueRequest
from authlete.dto.userinfo_request                import UserInfoRequest


class UserInfoRequestHandler(BaseRequestHandler):
    """Handler for userinfo requests to a userinfo endpoint.
    """


    def __init__(self, api, spi):
        """Constructor

        Args:
            api (authlete.api.AuthleteApi)
            spi (authlete.django.handler.spi.UserInfoRequestHandlerSpi)
        """

        super().__init__(api)
        self._spi = spi


    def handle(self, request):
        """Handle a userinfo request.

        This method calls Authlete's /auth/userinfo API and conditionally
        /auth/userinfo/issue API.

        Args:
            request (django.http.HttpRequest)

        Returns:
            django.http.HttpResponse

        Raises:
            authlete.api.AuthleteApiException
        """

        # Extract the access token from the request.
        accessToken = RequestUtility.extractAccessToken(request)

        if accessToken is None:
            # 400 Bad Request with a WWW-Authenticate header.
            return ResponseUtility.wwwAuthenticate(400,
                'Bearer error="invalid_token",error_description="An access token is required."')

        # Call Authlete's /api/auth/userinfo API.
        res = self.__callUserInfoApi(request, accessToken)

        # 'action' in the response denotes the next action which the
        # implementation of the userinfo endpoint should take.
        action = res.action

        # The content of the response to the client application.
        content = res.responseContent

        # Additional HTTP headers.
        headers = self.__prepareHeaders(res)

        if action == UserInfoAction.INTERNAL_SERVER_ERROR:
            # 500 Internal Server Error
            return ResponseUtility.wwwAuthenticate(500, content, headers)
        elif action == UserInfoAction.BAD_REQUEST:
            # 400 Bad Request
            return ResponseUtility.wwwAuthenticate(400, content, headers)
        elif action == UserInfoAction.UNAUTHORIZED:
            # 401 Unauthorized
            return ResponseUtility.wwwAuthenticate(401, content, headers)
        elif action == UserInfoAction.FORBIDDEN:
            # 403 Forbidden
            return ResponseUtility.wwwAuthenticate(403, content, headers)
        elif action == UserInfoAction.OK:
            # Return the user information.
            return self.__getUserInfo(res, headers)
        else:
            # 500 Internal Server Error
            # /auth/userinfo API returns an unknown action.
            return self.unknownAction('/auth/userinfo')


    def __callUserInfoApi(self, request, accessToken):
        req = UserInfoRequest()

        # The access token.
        req.token = accessToken

        # The request may contain a client certificate.
        req.clientCertificate = RequestUtility.extractClientCert(request)

        # The request may contain a DPoP proof JWT.
        req.dpop = request.headers.get('DPoP')

        # Call /api/auth/userinfo API.
        return self.api.userinfo(req)


    def __prepareHeaders(self, res):
        if res.dpopNonce is not None:
            return { 'DPoP-Nonce': res.dpopNonce }

        return None


    def __getUserInfo(self, response, headers):
        # Collect information about the user.
        claims = ClaimCollector(
            response.subject, response.claims, None, self._spi).collect()

        # The value of the 'sub' claim (optional)
        sub = self._spi.getSub()

        # Generate a response from the userinfo endpoint.
        return self.__userInfoIssue(response.token, claims, sub, headers)


    def __userInfoIssue(self, token, claims, sub, headers):
        # Call Authlete's /api/auth/userinfo/issue API.
        res = self.__callUserInfoIssueApi(token, claims, sub)

        # 'action' in the response denotes the next action which the
        # implementation of the userinfo endpoint should take.
        action = res.action

        # The content of the response to the client application.
        content = res.responseContent

        if action == UserInfoIssueAction.INTERNAL_SERVER_ERROR:
            # 500 Internal Server Error
            return ResponseUtility.wwwAuthenticate(500, content, headers)
        elif action == UserInfoIssueAction.BAD_REQUEST:
            # 400 Bad Request
            return ResponseUtility.wwwAuthenticate(400, content, headers)
        elif action == UserInfoIssueAction.UNAUTHORIZED:
            # 401 Unauthorized
            return ResponseUtility.wwwAuthenticate(401, content, headers)
        elif action == UserInfoIssueAction.FORBIDDEN:
            # 403 Forbidden
            return ResponseUtility.wwwAuthenticate(403, content, headers)
        elif action == UserInfoIssueAction.JSON:
            # 200 OK, application/json; charset=UTF-8
            return ResponseUtility.okJson(content, headers)
        elif action == UserInfoIssueAction.JWT:
            # 200 OK, application/jwt
            return ResponseUtility.okJwt(content, headers)
        else:
            # 500 Internal Server Error
            # /auth/userinfo/issue API returns an unknown action.
            return self.unknownAction('/auth/userinfo/issue')


    def __callUserInfoIssueApi(self, token, claims, sub):
        # Prepare a request for /api/auth/userinfo/issue API.
        req = UserInfoIssueRequest()
        req.token = token
        req.sub   = sub

        if claims is not None:
            req.claims = dumps(claims)

        # Call /api/auth/userinfo/issue API.
        return self.api.userinfoIssue(req)



#
# Copyright (C) 2019-2024 Authlete, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
# either express or implied. See the License for the specific
# language governing permissions and limitations under the
# License.


from .authorization_request_base_handler             import AuthorizationRequestBaseHandler
from .authorization_request_decision_handler         import AuthorizationRequestDecisionHandler
from .authorization_request_error_handler            import AuthorizationRequestErrorHandler
from .base_request_handler                           import BaseRequestHandler
from .claim_collector                                import ClaimCollector
from .configuration_request_handler                  import ConfigurationRequestHandler
from .credential_issuer_jwks_request_handler         import CredentialIssuerJwksRequestHandler
from .credential_issuer_metadata_request_handler     import CredentialIssuerMetadataRequestHandler
from .credential_jwt_issuer_metadata_request_handler import CredentialJwtIssuerMetadataRequestHandler
from .federation_configuration_request_handler       import FederationConfigurationRequestHandler
from .federation_registration_request_handler        import FederationRegistrationRequestHandler
from .introspection_request_handler                  import IntrospectionRequestHandler
from .jwks_request_handler                           import JwksRequestHandler
from .no_interaction_handler                         import NoInteractionHandler
from .par_request_handler                            import ParRequestHandler
from .revocation_request_handler                     import RevocationRequestHandler
from .token_request_base_handler                     import TokenRequestBaseHandler
from .token_request_handler                          import TokenRequestHandler
from .userinfo_request_handler                       import UserInfoRequestHandler



#
# Copyright (C) 2019-2024 Authlete, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
# either express or implied. See the License for the specific
# language governing permissions and limitations under the
# License.


import re
import urllib.parse
from django.conf import settings
from authlete.django.web.basic_credentials import BasicCredentials


class RequestUtility(object):
    BEARER_PATTERN = re.compile('^Bearer *(?P<parameter>[^ ]+) *$', re.I)
    DPOP_PATTERN   = re.compile('^DPoP *(?P<parameter>[^ ]+) *$', re.I)


    @classmethod
    def extractQueryString(cls, request):
        return request.META.get('QUERY_STRING')


    @classmethod
    def extractRequestBody(cls, request):
        body = request.body
        if body is None:
            return None

        encoding = request.encoding
        if encoding is None:
            encoding = settings.DEFAULT_CHARSET

        return body.decode(encoding)


    @classmethod
    def extractParameters(cls, request):
        if request.method == 'GET':
            return cls.extractQueryString(request)
        else:
            return cls.extractRequestBody(request)


    @classmethod
    def extractAuthorization(cls, request):
        return request.headers.get('Authorization')


    @classmethod
    def extractBearerToken(cls, request):
        # Get the value of the Authorization header.
        authorization = cls.extractAuthorization(request)
        if authorization is None:
            return None

        # Expecting the value matches "Bearer {token}"
        mo = cls.BEARER_PATTERN.match(authorization)

        # If the value does not match the pattern.
        if mo is None:
            return None

        return mo.group('parameter')


    @classmethod
    def extractDpopToken(cls, request):
        # Get the value of the Authorization header.
        authorization = cls.extractAuthorization(request)
        if authorization is None:
            return None

        # Expecting the value matches "DPoP {token}"
        mo = cls.DPOP_PATTERN.match(authorization)

        # If the value does not match the pattern.
        if mo is None:
            return None

        return mo.group('parameter')


    @classmethod
    def extractAccessToken(cls, request):
        # Try to extract a token from "Bearer {token}"
        accessToken = cls.extractBearerToken(request)
        if accessToken is not None:
            return accessToken

        # Try to extract a token from "DPoP {token}"
        accessToken = cls.extractDpopToken(request)
        if accessToken is not None:
            return accessToken

        # No access token is available.
        return None


    @classmethod
    def extractBasicCredentials(cls, request):
        return BasicCredentials.parse(request.headers.get('Authorization'))


    @classmethod
    def extractClientCert(cls, request):
        # RFC 9440 Client-Cert HTTP Header Field
        clientCert = request.headers.get('Client-Cert')

        if clientCert is not None:
            # The value of 'Client-Cert' should be 'sf-binary', which is
            # defined in RFC 8941 as follows.
            #
            #     sf-binary = ":" *(base64) ":"
            #     base64    = ALPHA / DIGIT / "+" / "/" / "="
            #

            # Remove the colons at the beginning and at the end.
            return clientCert[1:-1]

        # Try a well-known HTTP header, X-Ssl-Cert
        clientCert = request.headers.get('X-Ssl-Cert')

        if clientCert is None:
            return None

        # "(null)" is a value that misconfigured Apache servers will send
        # instead of a missing header. This happens when "SSLOptions" does
        # not include "+ExportCertData".
        if clientCert == '(null)':
            return None

        # Nginx's $ssl_client_escaped_cert holds a urlencoded client
        # certificate in the PEM format.
        if clientCert.startswith('-----BEGIN%20'):
            # URL-decode
            return urllib.parse.unquote(clientCert)

        return clientCert



#
# Copyright (C) 2019 Authlete, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
# either express or implied. See the License for the specific
# language governing permissions and limitations under the
# License.


from abc import ABCMeta, abstractmethod


class UserClaimProviderSpi(metaclass=ABCMeta):
    """Interface to get a claim value of a user.
    """


    @abstractmethod
    def getUserClaimValue(self, subject, claimName, languageTag):
        """Get the value of a claim of a user.

        This method may be called multiple times.

        Args:
            subject (str) : The subject (= unique identifier) of a user.
            claimName (str) : A claim name such as "name" and "family_name".
            languageTag (str) : A language tag such as "en" and "ja".

        Returns:
            object :
                The value of the claim. None if the value is not available.
                In most cases, an instance of str should be returned. When
                "claimName" is "address", an instance of authlete.dto.Address
                should be returned.
        """
        pass



#
# Copyright (C) 2019-2024 Authlete, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
# either express or implied. See the License for the specific
# language governing permissions and limitations under the
# License.


import setuptools


with open("README.md", "r") as fh:
    long_description = fh.read()


setuptools.setup(
    name="authlete-django",
    version="1.1.0",
    license="Apache License Version 2.0",
    author="Authlete, Inc.",
    author_email="admin@authlete.com",
    description="Django Library for Authlete",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/authlete/authlete-python-django",
    packages=[
        "authlete.django.handler",
        "authlete.django.handler.spi",
        "authlete.django.web",
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Topic :: Security",
    ],
    install_requires=[
        "authlete>=1.3.0",
    ]
)



#
# Copyright (C) 2019-2024 Authlete, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
# either express or implied. See the License for the specific
# language governing permissions and limitations under the
# License.


from authlete.django.handler.token_request_base_handler import TokenRequestBaseHandler
from authlete.django.web.request_utility                import RequestUtility
from authlete.django.web.response_utility               import ResponseUtility
from authlete.dto.token_action                          import TokenAction
from authlete.dto.token_fail_reason                     import TokenFailReason
from authlete.dto.token_request                         import TokenRequest


class TokenRequestHandler(TokenRequestBaseHandler):
    """Handler for token requests to a token endpoint.
    """


    def __init__(self, api, spi):
        """Constructor

        Args:
            api (authlete.api.AuthleteApi)
            spi (authlete.django.handler.spi.TokenRequestHandlerSpi)
        """

        super().__init__(api)
        self._spi = spi


    def handle(self, request):
        """Handle a token request.

        This method calls Authlete's /auth/token API and conditionally
        either /auth/token/issue API or /auth/token/fail API.

        Args:
            request (django.http.HttpRequest)

        Returns:
            django.http.HttpResponse

        Raises:
            authlete.api.AuthleteApiException
        """

        # Call Authlete's /auth/token API.
        res = self.__callTokenApi(request)

        # 'action' in the response denotes the next action which the
        # implementation of the token endpoint should take.
        action = res.action

        # The content of the response to the client application.
        content = res.responseContent

        # Additional HTTP headers.
        headers = self.__prepareHeaders(res)

        if action == TokenAction.INVALID_CLIENT:
            # 401 Unauthorized.
            return ResponseUtility.unauthorized(
                'Basic realm="token"', content, headers)
        elif action == TokenAction.INTERNAL_SERVER_ERROR:
            # 500 Internal Server Error
            return ResponseUtility.internalServerError(content, headers)
        elif action == TokenAction.BAD_REQUEST:
            # 400 Bad Request
            return ResponseUtility.badRequest(content, headers)
        elif action == TokenAction.PASSWORD:
            # Process the token request whose flow is "Resource OWner
            # Password Credentials".
            return self.__handlePassword(res, headers)
        elif action == TokenAction.OK:
            # 200 OK
            return ResponseUtility.okJson(content, headers)
        elif action == TokenAction.TOKEN_EXCHANGE:
            # Process the token exchange request (RFC 8693)
            return self.__handleTokenExchange(res, headers)
        elif action == TokenAction.JWT_BEARER:
            # Process the token request which uses the grant type
            # urn:ietf:params:oauth:grant-type:jwt-bearer (RFC 7523).
            return self.__handleJwtBearer(res, headers)
        elif action == TokenAction.ID_TOKEN_REISSUABLE:
            # The flow of the token request is the refresh token flow
            # and an ID token can be reissued.
            return self.__handleIdTokenReissuable(res, headers)
        else:
            # 500 Internal Server Error
            # /auth/token API returns an unknown action.
            return self.unknownAction('/auth/token')


    def __callTokenApi(self, request):
        req = TokenRequest()

        # The request parameters.
        req.parameters = RequestUtility.extractRequestBody(request) or ''

        # The request may contain the basic authentication for client_secret_basic.
        credentials      = RequestUtility.extractBasicCredentials(request)
        req.clientId     = credentials.userId
        req.clientSecret = credentials.password

        # The request may contain a client certificate.
        req.clientCertificate = RequestUtility.extractClientCert(request)

        # The request may contain a DPoP proof JWT.
        req.dpop = request.headers.get('DPoP')

        # Other parameters
        req.properties = self._spi.getProperties()

        # Call /api/auth/token API.
        return self.api.token(req)


    def __prepareHeaders(self, res):
        if res.dpopNonce is not None:
            return { 'DPoP-Nonce': res.dpopNonce }

        return None


    def __handlePassword(self, response, headers):
        # The ticket to call Authelte's /api/auth/token/* API.
        ticket = response.ticket

        # The credentials of the resource owner.
        username = response.username
        password = response.password

        # Validate the credentials.
        subject = self._spi.authenticateUser(username, password)

        # If the credentials of the resource owner are invalid.
        if subject is None:
            # The credentials are invalid. Nothing is issued.
            return self.tokenFail(
                ticket, TokenFailReason.INVALID_RESOURCE_OWNER_CREDENTIALS, headers)

        # Issue tokens.
        return self.tokenIssue(ticket, subject, self._spi.getProperties(), headers)


    def __handleTokenExchange(self, tokenResponse, headers):
        # Let the SPI implementation handle the token request.
        response = self._spi.tokenExchange(tokenResponse)

        # If the SPI implementation has prepared a token response, it is used.
        # Otherwise, a token response with "error":"unsupported_grant_type" is
        # returned.
        return self.__useOrUnsupported(response)


    def __handleJwtBearer(self, tokenResponse, headers):
        # Let the SPI implementation handle the token request.
        response = self._spi.jwtBearer(tokenResponse)

        # If the SPI implementation has prepared a token response, it is used.
        # Otherwise, a token response with "error":"unsupported_grant_type" is
        # returned.
        return self.__useOrUnsupported(response)


    def __handleIdTokenReissuable(self, tokenResponse, headers):
        # TODO: Support ID token reissuance.

        # Note that calling ResponseUtility.ok() here will result in that the
        # token endpoint behaves in the same way as before and no ID token is
        # returned.
        return ResponseUtility.ok(tokenResponse.responseContent, headers)


    def __useOrUnsupported(self, response):
        if response is not None:
            return response

        # Generate a token response that indicates that the grant type is not
        # supported.
        #
        #     400 Bad Request
        #     Content-Type: application/json
        #
        #     {"error":"unsupported_grant_type"}
        #
        return ResponseUtility.badRequest('{"error":"unsupported_grant_type"}')



#
# Copyright (C) 2024 Authlete, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
# either express or implied. See the License for the specific
# language governing permissions and limitations under the
# License.


from authlete.django.handler.base_request_handler import BaseRequestHandler
from authlete.django.web.request_utility          import RequestUtility
from authlete.django.web.response_utility         import ResponseUtility
from authlete.dto.pushed_auth_req_action          import PushedAuthReqAction
from authlete.dto.pushed_auth_req_request         import PushedAuthReqRequest


class ParRequestHandler(BaseRequestHandler):
    """Handler for PAR requests to a PAR endpoint (RFC 9126).
    """


    def __init__(self, api):
        """Constructor

        Args:
            api (authlete.api.AuthleteApi)
        """

        super().__init__(api)


    def handle(self, request):
        """Handle a PAR request.

        Args:
            request (django.http.HttpRequest)

        Returns:
            django.http.HttpResponse

        Raises:
            authlete.api.AuthleteApiException
        """

        # Call Authlete's /pushed_auth_req API.
        res = self.__callPushedAuthReqApi(request)

        # 'action' in the response denotes the next action which the
        # implementation of the PAR endpoint should take.
        action = res.action

        # The content of the response to the client application.
        content = res.responseContent

        # Additional HTTP headers.
        headers = self.__prepareHeaders(res)

        if action == PushedAuthReqAction.CREATED:
            # 201 Created
            return ResponseUtility.created(content, headers)
        elif action == PushedAuthReqAction.BAD_REQUEST:
            # 400 Bad Request
            return ResponseUtility.badRequest(content, headers)
        elif action == PushedAuthReqAction.UNAUTHORIZED:
            # 401 Unauthorized
            return ResponseUtility.unauthorized(
                'Basic realm="par"', content, headers)
        elif action == PushedAuthReqAction.FORBIDDEN:
            # 403 Forbidden
            return ResponseUtility.forbidden(content, headers)
        elif action == PushedAuthReqAction.PAYLOAD_TOO_LARGE:
            # 413 Too Large
            return ResponseUtility.tooLarge(content, headers)
        elif action == PushedAuthReqAction.INTERNAL_SERVER_ERROR:
            # 500 Internal Server Error
            return ResponseUtility.internalServerError(content, headers)
        else:
            # 500 Internal Server Error
            # /pushed_auth_req API returns an unknown action.
            return self.unknownAction('/pushed_auth_req')


    def __callPushedAuthReqApi(self, request):
        req = PushedAuthReqRequest();

        # The request parameters.
        req.parameters = RequestUtility.extractRequestBody(request) or ''

        # The request may contain the basic authentication for client_secret_basic.
        credentials      = RequestUtility.extractBasicCredentials(request)
        req.clientId     = credentials.userId
        req.clientSecret = credentials.password

        # The request may contain a client certificate.
        req.clientCertificate = RequestUtility.extractClientCert(request)

        # The request may contain a DPoP proof JWT.
        req.dpop = request.headers.get('DPoP')

        # Call Authlete's /pushed_auth_req API.
        return self.api.pushAuthorizationRequest(req)


    def __prepareHeaders(self, res):
        if res.dpopNonce is not None:
            return { 'DPoP-Nonce': res.dpopNonce }

        return None



#
# Copyright (C) 2019 Authlete, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
# either express or implied. See the License for the specific
# language governing permissions and limitations under the
# License.


from authlete.django.web.response_utility import ResponseUtility
from authlete.dto.introspection_action    import IntrospectionAction
from authlete.dto.introspection_request   import IntrospectionRequest


class AccessTokenValidator(object):
    def __init__(self, api):
        super().__init__()
        self._api = api
        self.__resetValidation()


    def __resetValidation(self):
        self._valid                  = False
        self._introspectionResponse  = None
        self._introspectionException = None
        self._errorResponse          = None


    @property
    def api(self):
        return self._api


    @property
    def valid(self):
        """Get the result of the access token validation.

        After a call of `validate()` method, this property holds the same value
        as was returned from `validate()`.

        Returns:
            bool : The result of the access token validation.
        """
        return self._valid


    @property
    def introspectionResponse(self):
        """Get the response from Authlete's /api/auth/introspection API.

        `validate()` method internally calls /api/auth/introspection API and sets
        the response from the API to this property. Note that this property
        remains `None` if the API call threw an exception, and in that error case,
        the `introspectionException` property is set.

        On entry of `validate()` method, this property is reset to `None`.

        Returns:
            authlete.dto.IntrospectionResponse
        """
        return self._introspectionResponse


    @property
    def introspectionException(self):
        """Get the exception raised by a call to Authlete's /api/auth/introspection API.

        `validate()` method internally calls Authlete's /api/auth/introspection
        API. If the API call threw an exception, the exception would be set to
        this property. Note that this property remains `None` if the API call
        succeeds, and in that successful case, the `introspectionResponse`
        property is set.

        On entry of `validate()` method, this property is reset to `None`.

        Returns:
            Exception
        """
        return self._introspectionException


    @property
    def errorResponse(self):
        """Get the error response that should be sent back to the client.

        This property is internally set by `validate()` method when `validate()`
        returns `False`. This error response complies with RFC 6750 (The OAuth
        2.0 Authorization Framework: Bearer Token Usage).

        On entry of `validate()` method, this property is reset to `None`.

        Returns:
            django.http.HttpResponse
        """
        return self._errorResponse


    def validate(self, accessToken, requiredScopes=None, requiredSubject=None):
        """Validate an access token.

        On entry, as the first step, the following properties are reset to
        False or None: `valid`, `introspectionResponse`, `introspectionException`
        and `errorResponse`.

        Then, this method internally calls Authlete's /api/auth/introspection
        API to get information about the access token.

        If the API call failed, the exception thrown by the API call is set to
        the `introspectionException` property and an error response
        (`500 Internal Server Error`) that should be returned to the client
        application is set to the `errorResponse` property. Then, this method
        set `False` to the `valid` property and returns `False`.

        If the API call succeeded, the response from the API is set to the
        `introspectionResponse` property. Then, this method checks the value
        of the `action` parameter in the response from the API.

        If the value of the `action` parameter is `OK`, this method sets `True`
        to the `valid` property and returns `True`.

        If the value of the `action` parameter is not `OK`, this method builds
        an error response that should be returned to the client application and
        sets it to the `errorResponse` property. Then, this method sets `False`
        to the `valid` property and returns `False`.

        If the given access token exists and has not expired, and optionally
        if the access token covers all the required scopes (in case
        `requiredScopes` was given) and the access token is associated with
        the required subject (in case `requiredSubject` was given), this method
        returns `True`. In other cases, this method returns `False`.

        Args:
            accessToken (str): An access token to be validated.
            requiredScopes (list of str):
                Scopes that the access token should have. If this parameter is not
                `None`, the implementation of Authlete's /api/auth/introspection
                API checks whether the access token covers all the required scopes.
                On the other hand, if `None` is given, Authlete does not conduct
                the validation on scopes.
            requiredSubject (str):
                Subject (= unique identifier of an end-user) that the access
                token should be associated with. If this parameter is not `None`,
                the implementation of Authlete's /api/auth/introspection API checks
                whether the access token is associated with the required subject.
                On the other hand, if `None` is given, Authlete does not conduct
                the validation on subject.

        Returns:
            bool: The result of access token validation.
        """

        # Reset properties that may have been set by the previous call.
        self.__resetValidation()

        try:
            # Call Authlete's /api/auth/introspection API.
            self._introspectionResponse = self.__callIntrospectionApi(
                accessToken, requiredScopes, requiredSubject)
        except Exception as cause:
            self._introspectionException = cause
            self._errorResponse          = self.__buildErrorFromException(cause)
            self._valid                  = False
            return False

        # The 'action' parameter in the response from /api/auth/introspection
        # denotes the next action that the API caller should take.
        action = self._introspectionResponse.action

        if action == IntrospectionAction.OK:
            # The access token is valid.
            self._valid = True
            return True
        else:
            self._errorResponse = self.__buildErrorFromResponse(self._introspectionResponse)
            self._valid         = False
            return False


    def __callIntrospectionApi(self, accessToken, requiredScopes, requiredSubject):
        # Prepare a request to /api/auth/introspection API.
        req = IntrospectionRequest()
        req.token   = accessToken
        req.scopes  = requiredScopes
        req.subject = requiredSubject

        # Call /api/auth/introspection API.
        return self.api.introspection(req)


    def __buildErrorFromException(self, cause):
        # The value for the WWW-Authenticate header.
        challenge = 'Bearer error="server_error",error_description="Introspection API call failed."'

        # Build a response that complies with RFC 6749.
        return ResponseUtility.wwwAuthenticate(500, challenge)


    def __buildErrorFromResponse(self, response):
        action = response.action

        if action == IntrospectionAction.INTERNAL_SERVER_ERROR:
            statusCode = 500
        elif action == IntrospectionAction.BAD_REQUEST:
            statusCode = 400
        elif action == IntrospectionAction.UNAUTHORIZED:
            statusCode = 401
        elif action == IntrospectionAction.FORBIDDEN:
            statusCode = 403
        else:
            statusCode = 500

        # In error cases, the 'responseContent' parameter in the response
        # from Authlete's /api/auth/introspection API contains a value for
        # the WWW-Authenticate header.
        challenge = response.responseContent

        # Build a response that complies with RFC 6749.
        return ResponseUtility.wwwAuthenticate(statusCode, challenge)



#
# Copyright (C) 2019 Authlete, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
# either express or implied. See the License for the specific
# language governing permissions and limitations under the
# License.


from authlete.django.web.response_utility import ResponseUtility


class BaseRequestHandler(object):
    """The base class for request handlers.
    """


    def __init__(self, api):
        self._api = api


    @property
    def api(self):
        return self._api


    def unknownAction(self, apiPath):
        """Create a response indicating that an unknown action was received.

        Args:
            apiPath (str) : The path of an Authlete API.

        Returns:
            django.http.HttpResponse
        """

        content = \
            '{"error":"server_error","error_description":"Authlete\'s " + \
            "{} API returned an unknown action."}'.format(apiPath)

        return ResponseUtility.internalServerError(content)



#
# Copyright (C) 2019 Authlete, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
# either express or implied. See the License for the specific
# language governing permissions and limitations under the
# License.


from authlete.api.authlete_api_exception          import AuthleteApiException
from authlete.django.handler.base_request_handler import BaseRequestHandler
from authlete.django.web.response_utility         import ResponseUtility


class JwksRequestHandler(BaseRequestHandler):
    """Handler for requests to an endpoint that exposes JSON Web Key Set document (RFC 7517).

    An OpenID Provider is required to expose its JSON web Key Set document so
    that client applications can (1) verify signatures by the OpenID Provider
    and (2) encrypt their requests to the OpenID Provider. The URI of a JWK
    Set endpoint can be found as the value of the "jwks_uri" metadata which is
    defined in "3. OpenID Provider Metadata" of "OpenID Connect Discovery 1.0"
    """


    def __init__(self, api):
        """Constructor

        Args:
            api (authlete.api.AuthleteApi)
        """

        super().__init__(api)


    def handle(self, request, pretty=True):
        """Handle a request to a JWK Set document endpoint.

        This method calls Authlete's /api/service/jwks/get API.

        Args:
            request (django.http.HttpRequest)
            pretty (bool) : True to format the JWK Set document in pretty format.

        Returns:
            django.http.HttpResponse

        Raises:
            authlete.api.AuthleteApiException
        """

        cause = None

        try:
            # Call Authlete's /api/service/jwks/get API. The API returns the
            # JWK Set (RFC 7517) of the service. The second argument given
            # to getServiceJwks() is False not to include private keys.
            jwks = self.api.getServiceJwks(pretty, False)

            # If no JWK Set for the service is registered.
            if jwks is None or len(jwks) == 0:
                # 204 No Content.
                return ResponseUtility.noContent()

            # 200 OK, application/json;charset=UTF-8
            return ResponseUtility.okJson(jwks)
        except AuthleteApiException as e:
            cause = e

        if cause.response is None or cause.response.status_code != 302:
            # Something wrong happend.
            raise cause

        # The value of the Location header of the response from the Authlete API.
        location = cause.response.headers.get('Location')

        # 302 Found with a Location header.
        return ResponseUtility.location(location)
